
/* 
 * one line to give the Gauche-_gpgme and an idea of what it does.
 * Copyright (C) 2017 aaauuue 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * _gpgme.c
 */

#include "_gpgme.h"

/*
 * The following function is a dummy one; replace it for
 * your C function definitions.
 */

ScmObj test__gpgme( void )
{
     return SCM_MAKE_STR( "_gpgme is working" );
}

static ScmObj passphrase_cb_lambda = NULL;

#define NULL_SCM_MAKE_STR(cstr) ((cstr)==NULL ? SCM_MAKE_STR( "" ) : SCM_MAKE_STR((cstr)))

gpgme_error_t _passphrase_cb( void *opaque, const char *uid_hint,
                              const char *passphrase_info, int last_was_bad,
                              int fd )
{
     /* fprintf( stderr, ">> _passphrase_cb\n" ); */
     ScmObj oport =
         Scm_MakePortWithFd( SCM_MAKE_STR( "_gpgme_using_fd_port" ),
                             SCM_PORT_OUTPUT, fd, SCM_PORT_BUFFER_LINE, FALSE );
     ScmEvalPacket packet;

     if( Scm_Apply( passphrase_cb_lambda,
                    SCM_LIST5( SCM_OBJ( opaque ), NULL_SCM_MAKE_STR( uid_hint ),
                               NULL_SCM_MAKE_STR( passphrase_info ),
                               SCM_MAKE_INT( last_was_bad ), oport ),
                    &packet ) < 0 ) {
          fprintf( stderr, "[%s] %s\n",
                   SCM_STRING_CONST_CSTRING(
                       Scm_ConditionTypeName( packet.exception ) ),
                   SCM_STRING_CONST_CSTRING(
                       Scm_ConditionMessage( packet.exception ) ) );
     } else {
          /*
           * fprintf( stderr, "Success to apply. [%d]\n",
           *          SCM_INT_VALUE( packet.results[0] ) );
           */
          ;
     }

     return 0;
}

ScmObj _g_gpgme_set_passphrase_cb( ScmObj ctx, ScmObj cb_lambda )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_PROCEDUREP( cb_lambda ) ) {
          Scm_Error( "Need procedure or lambda." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );

     /* fprintf( stderr, "_g_gpgme_set_passphrase_cb\n" ); */
     passphrase_cb_lambda = cb_lambda;
     gpgme_set_passphrase_cb( p->obj, _passphrase_cb, NULL );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_check_version( ScmObj require )
{
     if( !SCM_STRINGP( require ) ) {
          Scm_Error( "Need string." );
     }

     char *str = Scm_GetString( SCM_STRING( require ) );
     char *arg = NULL;

     if( strcmp( str, "" ) != 0 ) {
          arg = str;
     }

     SCM_OBJ( SCM_MAKE_STR( gpgme_check_version( arg ) ) );
}

ScmObj _g_gpgme_new_context( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     gpgme_error_t      err;

     err = gpgme_new( &p->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_release_context( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );

     gpgme_release( p->obj );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_new_data( ScmObj data )
{
     if( !SCM_GPGME_DATA_P( data ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_DATA *p = SCM_GPGME_DATA( data );
     gpgme_error_t   err;

     err = gpgme_data_new( &p->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_release_data( ScmObj data )
{
     if( !SCM_GPGME_DATA_P( data ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_DATA *p = SCM_GPGME_DATA( data );

     gpgme_data_release( p->obj );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_new_data_from_file( ScmObj buf, const char *filename, int copy )
{
     if( !SCM_GPGME_DATA_P( buf ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_DATA *data = SCM_GPGME_DATA( buf );
     gpgme_error_t   err;

     err = gpgme_data_new_from_file( &( data->obj ), filename, copy );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_new_data_from_string( ScmObj buf, ScmObj string, int copy )
{
     if( !SCM_GPGME_DATA_P( buf ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_STRINGP( string ) ) {
          Scm_Error( "Need gauche string." );
     }
          
     Scm_GPGME_DATA *data    = SCM_GPGME_DATA( buf );
     char           *_string = NULL;
     size_t          len     = 0;
     gpgme_error_t   err;

     _string = Scm_GetString( SCM_STRING( string ) );
     len = strlen( _string );

     err = gpgme_data_new_from_mem( &data->obj, _string, len, copy );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_set_protocol( ScmObj ctx, int protocol )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     gpgme_error_t      err;

     err = gpgme_set_protocol( p->obj, protocol );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_set_armor( ScmObj ctx, int yesno )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );

     gpgme_set_armor( p->obj, yesno );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_set_textmode( ScmObj ctx, int yesno )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );

     gpgme_set_textmode( p->obj, yesno );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_op_keylist_start( ScmObj ctx, const char *pattern,
                                  int secret_only )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     gpgme_error_t      err;

     err = gpgme_op_keylist_start( p->obj, pattern, secret_only );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_keylist_next( ScmObj ctx, ScmObj key )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_GPGME_KEY_P( key ) ) {
          Scm_Error( "Need key object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_KEY *    _key = SCM_GPGME_KEY( key );
     gpgme_error_t      err;

     err = gpgme_op_keylist_next( p->obj, &_key->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_keylist_end( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     gpgme_error_t      err;

     err = gpgme_op_keylist_end( p->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_get_key( ScmObj ctx, const char *pattern, ScmObj key,
                         int secret_only )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_KEY *    _key = SCM_GPGME_KEY( key );
     gpgme_error_t      err;

     err = gpgme_get_key( p->obj, pattern, &_key->obj, secret_only );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_signers_clear( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );

     gpgme_signers_clear( p->obj );

     return ERR_RETURN( 0 );
}

ScmObj _g_gpgme_signers_add( ScmObj ctx, ScmObj key )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_GPGME_KEY_P( key ) ) {
          Scm_Error( "Need key object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_KEY *    _key = SCM_GPGME_KEY( key );
     gpgme_error_t      err;

     err = gpgme_signers_add( p->obj, _key->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_sign( ScmObj ctx, ScmObj in, ScmObj out, int sign_mode )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_GPGME_DATA_P( in ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_GPGME_DATA_P( out ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_DATA *   _in = SCM_GPGME_DATA( in );
     Scm_GPGME_DATA *   _out = SCM_GPGME_DATA( out );
     gpgme_error_t      err;

     err = gpgme_op_sign( p->obj, _in->obj, _out->obj, sign_mode );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_verify( ScmObj ctx, ScmObj sig, ScmObj signed_text,
                           ScmObj plaintext )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_GPGME_DATA_P( sig ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_GPGME_DATA_P( signed_text ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_DATA *   _sig = SCM_GPGME_DATA( sig );
     Scm_GPGME_DATA *   _signed_text = SCM_GPGME_DATA( signed_text );
     /* Scm_GPGME_DATA *   _plain_text = SCM_GPGME_DATA( plaintext ); */
     Scm_GPGME_DATA *   _plain_text = NULL;
     gpgme_error_t      err;

     if( SCM_GPGME_DATA_P( plaintext ) ) {
          _plain_text = SCM_GPGME_DATA( plaintext );
          err = gpgme_op_verify( p->obj, _sig->obj, _signed_text->obj,
                                 _plain_text->obj );
     } else { /* detached sign */
          err = gpgme_op_verify( p->obj, _sig->obj, _signed_text->obj, NULL );
     }

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_verify_result( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *      p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_VERIFY_RESULT *vresult = SCM_GPGME_VERIFY_RESULT(
         gpgme_verify_result_allocate( NULL, SCM_MAKE_STR( "" ) ) );

     vresult->obj = gpgme_op_verify_result( p->obj );

     return SCM_OBJ( vresult );
}

ScmObj _g_gpgme_op_export_keys( ScmObj ctx, ScmObj keys, int export_mode, ScmObj buf )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_LISTP( keys ) ) {
          Scm_Error( "Need list object." );
     }

     if( !SCM_GPGME_DATA_P( buf ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_CONTEXT *p        = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_DATA    *_buf     = SCM_GPGME_DATA( buf );
     size_t             keys_len = 0;
     ScmObj             _p;
     gpgme_error_t      err;
     gpgme_key_t       *_keys    = NULL;
     int i = 0;

     keys_len = Scm_Length( keys );
     if( keys_len != 0 ) {
          _keys = ( gpgme_key_t * )malloc( sizeof( gpgme_key_t ) * ( keys_len + 1 ) );
          if( _keys == NULL ) {
               return ERR_RETURN( -1 );
          }

          SCM_FOR_EACH( _p, keys ) {
               _keys[i] = SCM_GPGME_KEY( SCM_CAR( _p ) )->obj;
               i++;
          }
          _keys[keys_len] = NULL;
     } else if( keys_len == 0 ) {
          Scm_Error( "Need export key more than one it." );
          return ERR_RETURN( 0 );
     }

     err = gpgme_op_export_keys( p->obj, _keys, export_mode, _buf->obj );

     if( _keys ) {
          free( _keys );
          _keys = NULL;
     }

     return ERR_RETURN(err);
}

long long _g_gpgme_data_seek( ScmObj data, long long offset, int whence )
{
     Scm_GPGME_DATA *_data = SCM_GPGME_DATA( data );
     long long       pos = 0;

     pos = gpgme_data_seek( _data->obj, offset, whence );

     return pos;
}

long _g_gpgme_data_read( ScmObj data, ScmObj buffer, size_t size )
{
     if( !SCM_GPGME_DATA_P( data ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_U8VECTORP( buffer ) ) {
          Scm_Error( "Need u8vector." );
     }

     Scm_GPGME_DATA *_data = SCM_GPGME_DATA( data );
     unsigned char * vec_ref;
     char *          buf = NULL;
     long            ret = 0;
     size_t          i = 0;

     buf = ( char * )malloc( sizeof( char ) * size );
     if( buf == NULL ) {
          return -1;
     }
     memset( buf, 0, size );
     vec_ref = SCM_U8VECTOR_ELEMENTS( buffer );

     ret = gpgme_data_read( _data->obj, buf, size );
     if( ret <= 0 ) {
          free( buf );
          buf = NULL;
          return 0;
     }

     for( i = 0; i < ret; ++i ) {
          vec_ref[i] = ( unsigned char )buf[i];
     }
     if( ret != size ) {
          vec_ref[ret] = 0;
     }
     free( buf );
     buf = NULL;

     return ret;
}

ScmObj _g_gpgme_op_encrypt( ScmObj ctx, ScmObj recp, int flags, ScmObj plain,
                            ScmObj cipher )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_LISTP( recp ) ) {
          Scm_Error( "Need list object." );
     }

     if( !SCM_GPGME_DATA_P( plain ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_GPGME_DATA_P( cipher ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_CONTEXT *p        = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_DATA *   _plain   = SCM_GPGME_DATA( plain );
     Scm_GPGME_DATA *   _cipher  = SCM_GPGME_DATA( cipher );
     size_t             recp_len = 0;
     ScmObj             _p;
     gpgme_error_t      err;
     gpgme_key_t *      _recp    = NULL;
     int                i        = 0;

     recp_len = Scm_Length( recp );
     if( recp_len != 0 ) {
          _recp = ( gpgme_key_t * )malloc( sizeof( gpgme_key_t ) * ( recp_len + 1 ) );
          if( _recp == NULL ) {
               return ERR_RETURN( -1 );
          }

          SCM_FOR_EACH( _p, recp ) {
               _recp[i] = SCM_GPGME_KEY( SCM_CAR( _p ) )->obj;
               i++;
          }
          _recp[recp_len] = NULL;
          err = gpgme_op_encrypt( p->obj, _recp, flags, _plain->obj, _cipher->obj );
     } else if( recp_len == 0 ) { /* symmetric encryption */
          err = gpgme_op_encrypt( p->obj, NULL, flags, _plain->obj, _cipher->obj );
     }
          
     if( _recp ) {
          free( _recp );
          _recp = NULL;
     }

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_op_decrypt( ScmObj ctx, ScmObj cipher, ScmObj plain )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     if( !SCM_GPGME_DATA_P( plain ) ) {
          Scm_Error( "Need data object." );
     }

     if( !SCM_GPGME_DATA_P( cipher ) ) {
          Scm_Error( "Need data object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     Scm_GPGME_DATA *   _cipher = SCM_GPGME_DATA( cipher );
     Scm_GPGME_DATA *   _plain = SCM_GPGME_DATA( plain );
     gpgme_error_t      err;

     err = gpgme_op_decrypt( p->obj, _cipher->obj, _plain->obj );

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_set_pinentry_mode( ScmObj ctx, int mode )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     gpgme_error_t      err;

     if( GPGME_VERSION_NUMBER >= 0x010400  ) {
          err = gpgme_set_pinentry_mode( p->obj, mode );
     } else {
          err = 0;
     }

     return ERR_RETURN( err );
}

ScmObj _g_gpgme_get_pinentry_mode( ScmObj ctx )
{
     if( !SCM_GPGME_CONTEXT_P( ctx ) ) {
          Scm_Error( "Need context object." );
     }

     Scm_GPGME_CONTEXT *p = SCM_GPGME_CONTEXT( ctx );
     int mode = GPGME_PINENTRY_MODE_DEFAULT;

     if( GPGME_VERSION_NUMBER >= 0x010400  ) {
          mode = gpgme_get_pinentry_mode( p->obj );
     }

     return SCM_MAKE_INT( mode );
}


/*
 * Module initialization function.
 */
extern void Scm_Init__gpgmelib( ScmModule * );

void Scm_Init__gpgme( void )
{
     ScmModule *mod;

     /* Register this DSO to Gauche */
     SCM_INIT_EXTENSION( _gpgme );

     /* Create the module if it doesn't exist yet. */
     mod = SCM_MODULE( SCM_FIND_MODULE( "_gpgme", TRUE ) );

     Scm_InitStaticClass( SCM_CLASS_GPGME_CONTEXT, "<gpgme-context>", mod,
                          gpgme_context_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_DATA, "<gpgme-data>", mod,
                          gpgme_data_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_ENGINE_INFO, "<gpgme-engine-info>",
                          mod, gpgme_engine_info_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_INVALID_KEY, "<gpgme-invalid-key>",
                          mod, gpgme_invalid_key_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_KEY, "<gpgme-key>", mod,
                          gpgme_key_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_KEY_SIG, "<gpgme-key-sig>", mod,
                          gpgme_key_sig_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_NEW_SIGNATURE,
                          "<gpgme-new-signature>", mod,
                          gpgme_new_signature_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_SIG_NOTATION, "<gpgme-sig-notation>",
                          mod, gpgme_sig_notation_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_SIGN_RESULT, "<gpgme-sign-result>",
                          mod, gpgme_sign_result_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_SIGNATURE, "<gpgme-signature>", mod,
                          gpgme_signature_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_SUBKEY, "<gpgme-subkey>", mod,
                          gpgme_subkey_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_TOFU_INFO, "<gpgme-tofu-info>", mod,
                          gpgme_tofu_info_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_USER_ID, "<gpgme-user-id>", mod,
                          gpgme_user_id_slots, 0 );
     Scm_InitStaticClass( SCM_CLASS_GPGME_VERIFY_RESULT,
                          "<gpgme-verify-result>", mod,
                          gpgme_verify_result_slots, 0 );

     /* Register stub-generated procedures */
     Scm_Init__gpgmelib( mod );
}
