#ifndef __GPGMEUTIL_H__
#define __GPGMEUTIL_H__

/* 
 * one line to give the Gauche-_gpgme and an idea of what it does.
 * Copyright (C) 2017 aaauuue 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <gauche.h>

#include "config.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* #ifdef HAVE_STDLIB_H */

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif /* #ifdef HAVE_LOCALE_H */

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* #ifdef HAVE_ERRNO_H */

#define _FILE_OFFSET_BITS 64
#include <gpgme.h>

#define STR( X ) #X
#define XSTR( X ) STR( X )
#define L __LINE__

#ifdef HAVE_ERRNO_H
#define fail_if_err( err )                                                     \
     do {                                                                      \
          if( err ) {                                                          \
               fprintf( stderr, "%s:%d: %s: %s\n", __FILE__, __LINE__,         \
                        gpgme_strsource( err ), gpgme_strerror( err ) );       \
          }                                                                    \
     } while( 0 )
#else
#define fail_if_err( err )                                                     \
     do {                                                                      \
          if( err ) {                                                          \
               fprintf( stderr, "%s:%d: %s\n", __FILE__, __LINE__,             \
                        gpgme_strsource( err ) );                              \
          }                                                                    \
     } while( 0 )

#endif /* #ifdef HAVE_ERRNO_H */

#ifdef HAVE_ERRNO_H
#define ERR_RETURN( err )                                                      \
     Scm_Values4( SCM_MAKE_INT( err ), SCM_MAKE_STR( __FILE__ ":" XSTR( L ) ), \
                  SCM_MAKE_STR( gpgme_strsource( err ) ),                      \
                  SCM_MAKE_STR( gpgme_strerror( err ) ) )
#else
     Scm_Values4( SCM_MAKE_INF( err ), SCM_MAKE_STR( __FILE__ ":" XSTR( L ) ), \
                  SCM_MAKE_STR( gpgme_strsource( err ) ),
                  SCM_MAKE_STR( "error has occured." ) )
#endif /* #ifdef HAVE_ERRNO_H */

void init_gpgme( void );

/* from gpgme test code */
void print_data( gpgme_data_t dh );
char *nullout( char *src );
void print_key( gpgme_key_t key );

#endif /* __GPGMEUTIL_H__ */
