#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(use text.csv)
(use srfi-13)

(define csv-reader (make-csv-reader #\,))

(define *def-source* '())
(define *includes* '())
(define *gpl* "")

(define (read-gpl)
  (call-with-input-file "gpl.txt"
    (^p
     (set! *gpl* (open-output-string))
     (port-for-each
      (^c 
       (format *gpl* " * ~a\n" c))
      (cut read-line p)))))

;; (read-gpl)

;; (display 
;;  (get-output-string *gpl*)
;;  )

(define (build-def-source src)
  (set! *def-source* '())
  (set! *includes* '())
  (call-with-input-file src 
    (^i
     (port-for-each
      ;; (^c (format #t "~S\n" c))
      (^c
       (if (not (rxmatch #/^#.*/ (~ c 0)))
         (set! *def-source* (append *def-source* (list c)))))
      (cut csv-reader i)))))

(define (gen-header prefix :optional (oport (current-output-port)))
  (define (gen-declare)
    (format oport
            "ScmObj ~a_allocate( ScmClass *klass, ScmObj initargs );

"
            prefix)

    (dolist (call *def-source*)
      (if (not (string=? (~ call 0) "r"))
        (format oport "ScmObj ~a_~a( Scm_~a *p );\n"
                prefix (~ call 2) (string-upcase prefix)))))

  (define (gen-slot-spec)
    (format oport "\nstatic ScmClassStaticSlotSpec ~a_slots[] = {\n"
            prefix)
    (dolist (call *def-source*)
      (if (not (string=? (~ call 0) "r"))
        (format oport "SCM_CLASS_SLOT_SPEC( \"~a\", ~a_~a, NULL ),\n"
                #0=(~ call 2) prefix #0#)))
    (format oport "SCM_CLASS_SLOT_SPEC_END()\n};\n"))

  (define (gen-struct)
    (format oport "typedef struct Scm_~aRec {\nSCM_HEADER;\n"
            (string-upcase prefix))

    (dolist (call *def-source*)
      (if (string=? (~ call 0) "r")
        (format oport "~a ~a;\n" (~ call 1) (~ call 2))))

    (format oport "} Scm_~a;\n\nSCM_CLASS_DECL( Scm_~aClass );
#define SCM_CLASS_~a (&Scm_~aClass)
#define SCM_~a(obj) ((Scm_~a*)obj)
#define SCM_~a_P(obj) SCM_XTYPEP(obj, &Scm_~aClass)

"
            #2=(string-upcase prefix) #2# #2# #2# #2# #2# #2# #2#))
  

  (format oport "

#include <gauche.h>
#include <gauche/extend.h>
#include <gauche/class.h>

#include \"config.h\"
#include \"gpgmeutil.h\"

")

  (dolist (include *includes*)
    (format oport include))
  (format oport "\n\n")

  (gen-struct)
  (gen-declare)
  (gen-slot-spec)

  )

(define (gen-body prefix :optional (oport (current-output-port)))
  (define (new-name src)
    (let ([m (rxmatch #/(.*)_t$/ src)])
      (if m
        (rxmatch-substring m 1)
        src)))

  (define (re-type src call origin)
    (cond
     [(string=? src "s") (format "if( p->obj == NULL ) { return SCM_FALSE; }
if( p->obj->~a == NULL ) {
return SCM_FALSE;
} else {
return SCM_MAKE_STR( p->obj->~a );
}
" call call)]
     [(string=? src "i") (format "if( p->obj == NULL ) { return SCM_FALSE; }

return SCM_MAKE_INT( p->obj->~a );" call)]
     [(string=? src "o") (begin
                           (if (not (string=? (new-name (~ origin 1)) prefix))
                             (set! *includes*
                                   (append
                                    *includes*
                                    (list
                                     (format "#include \"struct_~a.h\"\n"
                                             (new-name (~ origin 1)))))))
                           (format "if( p->obj == NULL ) { return SCM_FALSE; }

if( p->obj->~a == NULL ) {
return SCM_FALSE;
}
Scm_~a *q = SCM_~a( ~a_allocate( NULL, SCM_MAKE_STR( \"\" ) ) );
q->obj = p->obj->~a;
return SCM_OBJ( q );
" call (string-upcase (new-name (~ origin 1)))
(string-upcase (new-name (~ origin 1))) (new-name (~ origin 1)) call))]
     [else ""]))

  (define (gen-getter)
    (let ([s (open-output-string)])
      (dolist (call *def-source*)
        (if (not (string=? (~ call 0) "r"))
          (format s "ScmObj ~a_~a( Scm_~a *p )
{
~a
}

"
        prefix (~ call 2) (string-upcase prefix) (re-type (~ call 0) (~ call 2) call))))

(get-output-string s)))

(define (gen-init)
  (let ((s (open-output-string)))
    (dolist (call *def-source*)
      (if (string=? (~ call 0) "r")
        (format s "p->~a = NULL;" (~ call 2))))
    (get-output-string s)))

(let ([name (string-upcase prefix)])
  (format oport "

ScmObj ~a_allocate( ScmClass *klass, ScmObj initargs )
{
 Scm_~a *p = SCM_NEW( Scm_~a );
        SCM_SET_CLASS( p, SCM_CLASS_~a );

        ~a

        return SCM_OBJ( p );
        }

SCM_DEFINE_BUILTIN_CLASS( Scm_~aClass,
                          NULL, NULL, NULL,
                          ~a_allocate, NULL );

~a

"
prefix name name name (gen-init) (string-upcase prefix) prefix (gen-getter))))

(define (main args) 
  (if (< (length args) 4)
    (begin
      (format (current-error-port) "~a out-file-name prefix def-file\n"
              (~ args 0))
      (sys-exit 1)))
  
  (read-gpl)
  (build-def-source (~ args 3))
  
  (call-with-output-file (format "~a.c" (~ args 1))
    (^p
     (format p "\n/*\n~a */\n" (get-output-string *gpl*))
     (format p "\n#include \"~a.h\"\n\n" (~ args 1))
     (gen-body (~ args 2) p)))

(call-with-output-file (format "~a.h" (~ args 1))
    (^p
     (format p "#ifndef __~a_H__\n#define __~a_H__\n\n"
             (string-upcase (~ args 1)) (string-upcase (~ args 1)))
     (format p "/*\n~a */" (get-output-string *gpl*))
     (gen-header (~ args 2) p)
     (format p "#endif /* #ifndef __~a_H__ */\n"
             (string-upcase (~ args 1)))))

  0)
