#ifndef __STRUCT_GPGME_KEY_H__
#define __STRUCT_GPGME_KEY_H__

/*
 * one line to give the Gauche-_gpgme and an idea of what it does.
 * Copyright (C) 2017 aaauuue
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <gauche.h>
#include <gauche/extend.h>
#include <gauche/class.h>

#include "config.h"
#include "gpgmeutil.h"

#include "struct_gpgme_subkey.h"
#include "struct_gpgme_user_id.h"
#include "struct_gpgme_subkey.h"
#include "struct_gpgme_user_id.h"

typedef struct Scm_GPGME_KEYRec {
     SCM_HEADER;
     gpgme_key_t obj;
} Scm_GPGME_KEY;

SCM_CLASS_DECL( Scm_GPGME_KEYClass );
#define SCM_CLASS_GPGME_KEY ( &Scm_GPGME_KEYClass )
#define SCM_GPGME_KEY( obj ) ( ( Scm_GPGME_KEY * )obj )
#define SCM_GPGME_KEY_P( obj ) SCM_XTYPEP( obj, &Scm_GPGME_KEYClass )

ScmObj gpgme_key_allocate( ScmClass *klass, ScmObj initargs );

ScmObj gpgme_key_revoked( Scm_GPGME_KEY *p );
ScmObj gpgme_key_expired( Scm_GPGME_KEY *p );
ScmObj gpgme_key_disabled( Scm_GPGME_KEY *p );
ScmObj gpgme_key_invalid( Scm_GPGME_KEY *p );
ScmObj gpgme_key_can_encrypt( Scm_GPGME_KEY *p );
ScmObj gpgme_key_can_sign( Scm_GPGME_KEY *p );
ScmObj gpgme_key_can_certify( Scm_GPGME_KEY *p );
ScmObj gpgme_key_secret( Scm_GPGME_KEY *p );
ScmObj gpgme_key_can_authenticate( Scm_GPGME_KEY *p );
ScmObj gpgme_key_is_qualified( Scm_GPGME_KEY *p );
ScmObj gpgme_key_origin( Scm_GPGME_KEY *p );
ScmObj gpgme_key_protocol( Scm_GPGME_KEY *p );
ScmObj gpgme_key_issuer_serial( Scm_GPGME_KEY *p );
ScmObj gpgme_key_issuer_name( Scm_GPGME_KEY *p );
ScmObj gpgme_key_chain_id( Scm_GPGME_KEY *p );
ScmObj gpgme_key_owner_trust( Scm_GPGME_KEY *p );
ScmObj gpgme_key_subkeys( Scm_GPGME_KEY *p );
ScmObj gpgme_key_uids( Scm_GPGME_KEY *p );
ScmObj gpgme_key__last_subkey( Scm_GPGME_KEY *p );
ScmObj gpgme_key__last_uid( Scm_GPGME_KEY *p );
ScmObj gpgme_key_keylist_mode( Scm_GPGME_KEY *p );
ScmObj gpgme_key_fpr( Scm_GPGME_KEY *p );
ScmObj gpgme_key_last_update( Scm_GPGME_KEY *p );

static ScmClassStaticSlotSpec gpgme_key_slots[] = {
     SCM_CLASS_SLOT_SPEC( "revoked", gpgme_key_revoked, NULL ),
     SCM_CLASS_SLOT_SPEC( "expired", gpgme_key_expired, NULL ),
     SCM_CLASS_SLOT_SPEC( "disabled", gpgme_key_disabled, NULL ),
     SCM_CLASS_SLOT_SPEC( "invalid", gpgme_key_invalid, NULL ),
     SCM_CLASS_SLOT_SPEC( "can_encrypt", gpgme_key_can_encrypt, NULL ),
     SCM_CLASS_SLOT_SPEC( "can_sign", gpgme_key_can_sign, NULL ),
     SCM_CLASS_SLOT_SPEC( "can_certify", gpgme_key_can_certify, NULL ),
     SCM_CLASS_SLOT_SPEC( "secret", gpgme_key_secret, NULL ),
     SCM_CLASS_SLOT_SPEC( "can_authenticate", gpgme_key_can_authenticate,
                          NULL ),
     SCM_CLASS_SLOT_SPEC( "is_qualified", gpgme_key_is_qualified, NULL ),
     SCM_CLASS_SLOT_SPEC( "origin", gpgme_key_origin, NULL ),
     SCM_CLASS_SLOT_SPEC( "protocol", gpgme_key_protocol, NULL ),
     SCM_CLASS_SLOT_SPEC( "issuer_serial", gpgme_key_issuer_serial, NULL ),
     SCM_CLASS_SLOT_SPEC( "issuer_name", gpgme_key_issuer_name, NULL ),
     SCM_CLASS_SLOT_SPEC( "chain_id", gpgme_key_chain_id, NULL ),
     SCM_CLASS_SLOT_SPEC( "owner_trust", gpgme_key_owner_trust, NULL ),
     SCM_CLASS_SLOT_SPEC( "subkeys", gpgme_key_subkeys, NULL ),
     SCM_CLASS_SLOT_SPEC( "uids", gpgme_key_uids, NULL ),
     SCM_CLASS_SLOT_SPEC( "_last_subkey", gpgme_key__last_subkey, NULL ),
     SCM_CLASS_SLOT_SPEC( "_last_uid", gpgme_key__last_uid, NULL ),
     SCM_CLASS_SLOT_SPEC( "keylist_mode", gpgme_key_keylist_mode, NULL ),
     SCM_CLASS_SLOT_SPEC( "fpr", gpgme_key_fpr, NULL ),
     SCM_CLASS_SLOT_SPEC( "last_update", gpgme_key_last_update, NULL ),
     SCM_CLASS_SLOT_SPEC_END()
};
#endif /* #ifndef __STRUCT_GPGME_KEY_H__ */
