#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(use parser.peg)
(use srfi-13)
(use util.match)
(use file.util)

(define (arg-splitter src)
  (string-split (string-trim-both src) ", "))

(define %ws ($skip-many ($one-of #[ \t\r\n])))
(define %extern ($s "extern"))
(define %types ($or ($s "int") ($s "unsigned int")
                    ($s "long long") ($s "void")
                    ($s "const char")
                    ($s "long") ($s "size_t")
                    ($s "gpgme_protocol_t")
                    ($s "ScmObj")))
(define %name ($many-till anychar ($c #\()))
(define %args ($between ($string "(") ($many-till anychar ($string ")")) ($string ")")))

(define %declare
  ($do
   (_ %extern)
   (_ %ws)
   (types %types)
   (_ %ws)
   (name %name)
   (args %args)
   ;; (_ ($string ";"))
   ($return (list types
                  (list->string name)
                  (arg-splitter (list->string args))))))

;; (peg-parse-string %declare "extern unsigned int _g_gpgme_new_context( ScmObj *ctx );")
;; (peg-parse-string %declare "extern unsigned int _g_gpgme_new_data_from_file( ScmObj *buf, ScmObj *filename, int copy );")

(define %arg-parse
  ($do
   (type %types)
   (_ %ws)
   (name ($many anychar))
   ($return (list type (list->string name)))))

;; (define z
;;   (string-split 
;;  (string-trim-both " ScmObj *buf, ScmObj *filename, int copy ")
;;  ", "))

;; (dolist (_z z)
;;   (print (peg-parse-string %arg-parse _z)))

(define (transform src)
  (let ([t1 (peg-parse-string %declare src)]
        [t2 '()])
    (dolist (arg (~ t1 2))
      (set! t2 (append t2 (list (peg-parse-string %arg-parse arg)))))
    (list (~ t1 0) (~ t1 1) t2)))

;; (transform "extern unsigned int _g_gpgme_new_data_from_file( ScmObj *buf, ScmObj *filename, int copy );")

(define (reformat lst)
  (define (name-clean src)
    (let ([m (rxmatch #/_g_(.*)/ src)])
      (if m (rxmatch-substring m 1) src)))

  (define (arg-clean src)
    (let ([m (rxmatch #/^\*(.*)/ src)])
      (if m (rxmatch-substring m 1) src)))
  
  (define (proc-ret src)
    (cond
     [(string=? src "void") "::<void>"]
     [(string=? src "unsigned int") "::<uint>"]
     [(string=? src "int") "::<int>"]
     [(string=? src "const char") "::<const-cstring>"]
     [(string=? src "long long") "::<integer>"]
     [(string=? src "long") "::<uint>"]
     [(string=? src "size_t") "::<uint>"]
     [(string=? src "gpgme_protocol_t") "::<int>"]
     [else ""]))

  (define (proc-args src)
    (string-join 
    (map (^o
          (format "~a~a"
                  (arg-clean (~ o 1))
                  (proc-ret (~ o 0))))
         src)
    " "))
  
  (match lst
    [(ret-type name args)
     (begin
       (format "(define-cproc %~a (~a)~a ~a)"
               (regexp-replace-all #/_/ (name-clean name) "-")
               (proc-args args) (proc-ret ret-type) name))]))

    
  
;; (reformat
;;  (transform "extern unsigned int _g_gpgme_new_data_from_file( ScmObj *buf, ScmObj *filename, int copy );")
;;  )

(define (main args) 
  (dolist (line (file->string-list (~ args 1)))
    ($ print $ reformat $ transform line))
  0)
